# -*- coding: utf-8 -*-
# @Time    : 2019-04-11 19:32
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: globals.py
# @Software: PyCharm
import redis

from dag_manager.app.config import REDIS_CONFIG

conn_pool = redis.ConnectionPool(max_connections=128, retry_on_timeout=True, socket_keepalive=True, **REDIS_CONFIG)


def get_redis_client():
    return redis.Redis(connection_pool=conn_pool)



# -*- coding: utf-8 -*-
# @Time    : 2019-04-09 11:02
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: app.py
# @Software: PyCharm
from flask import Flask, request, flash, redirect

from dag_manager.app.blueprints.dag_ids import dag_ids_blueprint
from dag_manager.app.blueprints.dags import dags_blueprint

app = Flask(__name__)
app.register_blueprint(dags_blueprint)
app.register_blueprint(dag_ids_blueprint)


doc = "<pre>{}</pre>".format('\n'.join([
    str(rule).replace('<', '&lt;').replace('>', '&gt;') + '\t' + str(rule.methods)
    for rule in app.url_map.iter_rules()
]))


@app.route('/')
def index():
    return doc


@app.route('/echo')
def echo():
    return 'ok'


def upload_file():
    if request.method == 'POST':
        # check if the post request has the file part
        if 'file' not in request.files:
            flash('No file part')
            return redirect(request.url)
        file = request.files['file']
        # if user does not select file, browser also
        # submit an empty part without filename
        if file.filename == '':
            flash('No selected file')
            return redirect(request.url)
    return '''
    <!doctype html>
    <title>Upload new File</title>
    <h1>Upload new File</h1>
    <form method=post enctype=multipart/form-data>
      <input type=file name=file>
      <input type=submit value=Upload>
    </form>
    '''


if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-
# @Time    : 2019-04-16 11:40
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: app.py
# @Software: PyCharm
import os
import shutil
from zipfile import ZipFile

from flask import Flask, request

from dag_receiver.config import DAG_FOLDER

app = Flask(__name__)


@app.route('/<dag_id>/<op>', methods=['POST'])
def index(dag_id, op):
    path = os.path.join(DAG_FOLDER, dag_id)
    print(path)
    if op == 'DELETE':
        if not os.path.exists(path) or not os.path.isdir(path):
            return 'dag[{}] not found'.format(dag_id), 500
        shutil.rmtree(path)
    else:
        if os.path.exists(path):
            shutil.rmtree(path)
        os.mkdir(path)
        file = request.files['file']
        print(file)
        file_path = os.path.join(path, file.filename)
        file.save(file_path)
        if file_path.endswith('.zip'):
            with ZipFile(file_path) as zf:
                zf.extractall(path)
    return 'ok'


@app.route('/echo')
def echo():
    return 'ok'


if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-
# @Time    : 2019-04-11 15:05
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: lock.py
# @Software: PyCharm
"""
实现分布式锁
"""
import functools
import time

from dag_manager.app.globals import get_redis_client


def try_lock(key):
    ok = get_redis_client().setnx(key, 1) == 1
    get_redis_client().expire(key, 600)
    return ok


def lock(key):
    while get_redis_client().setnx(key, 1) == 0:
        print('lock failed')
        time.sleep(0.2)
    get_redis_client().expire(key, 600)


def unlock(key):
    get_redis_client().delete(key)


def locking(key):
    def outer(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            lock(key)
            result = func(*args, **kwargs)
            unlock(key)
            return result
        return wrapper
    return outer



# -*- coding: utf-8 -*-
# @Time    : 2019-04-11 18:05
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: dag_ids.py
# @Software: PyCharm
import json
import re

from flask import Blueprint

from dag_manager.app.config import (
    REDIS_KEYS,
)
from dag_manager.app.globals import get_redis_client

dag_ids_blueprint = Blueprint('dag_ids', __name__, url_prefix='/dag_ids')


@dag_ids_blueprint.route('/<pattern>')
def get(pattern):
    if re.match(r'^[A-Za-z_*][A-Za-z0-9_\*]*$', pattern) is None:
        return 'invalid pattern', 500
    pattern = pattern.replace(r'*', r'[[A-Za-z0-9_]*')
    dag_ids = [dag_id.decode('utf-8') for dag_id in get_redis_client().smembers(REDIS_KEYS['all_dag_ids'])]
    return json.dumps(list(filter(lambda dag_id: re.match(pattern, dag_id) is not None, dag_ids)))



# -*- coding: utf-8 -*-
# @Time    : 2019-04-09 15:02
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: app.py
# @Software: PyCharm
import functools
import logging
import os
import shutil
import uuid
import zipfile
from datetime import datetime
from enum import Enum

import requests
from airflow.models import DAG
from flask import (
    Blueprint,
    request,
    redirect,
    url_for,
)
from werkzeug.utils import secure_filename

from dag_manager.app.auth import need_auth
from dag_manager.app.config import (
    ALL_DAGS_ZIP_NAME,
    DAG_FOLDER,
    ALL_DAGS_ZIP_PATH,
    ACCEPTED_EXT,
    REDIS_KEYS,
    SYNC_URLS,
)
from dag_manager.app.globals import get_redis_client
from dag_manager.app.util.lock import unlock, try_lock


class OP(Enum):
    ADD = 'ADD'
    DELETE = 'DELETE'
    UPDATE = 'UPDATE'


dags_blueprint = Blueprint('dags', __name__, url_prefix='/dags')


def not_zipping(func):
    """
    该装饰器，用以判断是否
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        key = REDIS_KEYS['lock_modification']
        if not try_lock(key):
            return 'dag modification locked now, please modify dag later.', 500
        result = func(*args, **kwargs)
        unlock(key)
        return result

    return wrapper


def last_update_time():
    return float(get_redis_client().get(REDIS_KEYS['last_update_time']) or 0)


def last_zip_time():
    return float(get_redis_client().get(REDIS_KEYS['last_zip_time']) or 0)


def set_last_update_time(time):
    get_redis_client().set(REDIS_KEYS['last_update_time'], time)


def set_last_zip_time(time):
    get_redis_client().set(REDIS_KEYS['last_zip_time'], time)


@dags_blueprint.route('/')
@need_auth
@not_zipping
def get_all_dags():
    if last_zip_time() < last_update_time() or not os.path.exists(ALL_DAGS_ZIP_PATH):
        zip_dags()
    return redirect(url_for('static', filename='all_dags.zip'))


@dags_blueprint.route('/<dag_id>')
@need_auth
@not_zipping
def get(dag_id):
    cur = find_dag(dag_id)
    if cur is None:
        return '{dag_id}.py or {dag_id}.zip not found'.format(dag_id=dag_id), 500
    return redirect(url_for('static', filename=os.path.basename(cur)))


@dags_blueprint.route('/<dag_id>', methods=['DELETE'])
@need_auth
@not_zipping
def delete(dag_id):
    cur = find_dag(dag_id)
    if cur is None:
        return '{dag_id}.py or {dag_id}.zip not found'.format(dag_id=dag_id)
    os.remove(cur)
    set_last_update_time(datetime.utcnow().timestamp())
    replicate(cur, dag_id, OP.DELETE)
    get_redis_client().srem(REDIS_KEYS['all_dag_ids'], dag_id)
    return 'done'


@dags_blueprint.route('/<dag_id>', methods=['POST'])
@need_auth
@not_zipping
def post(dag_id):
    """
    可上传的dag分为两类，1）单个py文件 2）包含依赖的zip文件，后者需要使用dag_path指定dag文件的路径
    todo 对zip是否需要更严格检查
    :param dag_id:
    :return:
    """
    file = request.files['file']
    _, ext = os.path.splitext(file.filename)
    if ext not in ACCEPTED_EXT:
        return 'not accept ext: {}'.format(ext), 500
    # 将文件存储在临时目录/tmp，并在其中检查Dag文件是否合法
    tmp_path = os.path.join('/tmp', secure_filename(file.filename))
    file.save(tmp_path)
    dag_path = request.form.get('dag_path')
    print('dag_path', dag_path)
    ok, reason = is_valid_dag_file(dag_id, tmp_path, ext, dag_path)
    if not ok:
        return reason, 500
    else:
        # 如果dag文件合法，则
        # 1）若dag已存在，先删除
        # 2）将文件从临时目录移动到 os.path.join(DAG_FOLDER, dag_id + ext)
        # 3）将更新同步至airflow scheduler、webserver结点
        # 4）记录更新时间
        # 5）将dag_id加入redis集合中
        # todo 保持操作atomic
        cur = find_dag(dag_id)
        if cur is not None:
            os.remove(cur)
        real_path = os.path.join(DAG_FOLDER, dag_id + ext)
        os.rename(tmp_path, real_path)
        replicate(real_path, dag_id, OP.UPDATE if get_redis_client().sismember(REDIS_KEYS['all_dag_ids'], dag_id) else OP.ADD)
        set_last_update_time(datetime.utcnow().timestamp())
        get_redis_client().sadd(REDIS_KEYS['all_dag_ids'], dag_id)
        return 'ok'


def zip_dags():
    with zipfile.ZipFile(ALL_DAGS_ZIP_PATH, mode='w') as zf:
        for f in os.listdir(DAG_FOLDER):
            if os.path.isfile(os.path.join(DAG_FOLDER, f)) and os.path.splitext(f)[1] in ACCEPTED_EXT and f != ALL_DAGS_ZIP_NAME:
                zf.write(os.path.join(DAG_FOLDER, f), f)
    set_last_zip_time(datetime.utcnow().timestamp())


def find_dag(dag_id):
    path = os.path.join(DAG_FOLDER, dag_id + '.py')
    if os.path.exists(path):
        return path
    path = os.path.join(DAG_FOLDER, dag_id + '.zip')
    if os.path.exists(path):
        return path
    return None


def is_valid_dag_file(dag_id, path, ext, dag_path=None):
    """
    合法需要符合以下条件：
    1、对于py文件，其包含一个模块级的Dag对象，且id与dag_id一样
    2、对于zip文件夹，仅检查dag_path指定的py文件，其包含一个模块级的Dag对象，且id与dag_id一样，且dag_path的文件名（除后缀）与dag_id一样
    # todo 更严格检查解压后的zip文件夹
    :param dag_id: dag_id
    :param path: 文件或者文件夹路径
    :param dag_path:
    :param ext: 后缀名
    :return: (is_valid, reason)
    """
    from importlib import util
    dag = None
    module_name = '_' + str(uuid.uuid1()).replace('-', '_')
    if ext == '.zip':
        if dag_path is None:
            return False, 'must specify dag_path where upload a zip file'
        zip_tract_path = tract_zip(path)
        path = os.path.join(zip_tract_path, dag_path)
    try:
        spec = util.spec_from_file_location(module_name, path)
        mod = util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        if ext == '.zip':
            shutil.rmtree(zip_tract_path)
    except Exception:
        return False, 'load dag exception'
    for v in dir(mod):
        if isinstance(getattr(mod, v), DAG):
            if dag is not None:
                return False, 'there are more than one dags'
            dag = getattr(mod, v)
    if dag is None:
        return False, 'dag named "{}" not found'.format(dag_id)
    if dag.dag_id != dag_id or \
            (ext == '.zip' and dag_id != os.path.basename(os.path.splitext(dag_path)[0])):
        return False, 'incorrect dag_id'
    return True, 'ok'


def tract_zip(filename):
    """
    将zip文件解压至/tmp/{uuid}中
    :param filename: 待解压文件名
    :return: /tmp/{uuid}
    """
    p = os.path.join('/tmp', str(uuid.uuid1()))
    with zipfile.ZipFile(filename) as zf:
        zf.extractall(p)
    return p


def replicate(path, dag_id, op):
    """
    将dag的更新复制给同步结点
    :param path: 文件路径
    :param dag_id:
    :param op: OP object
    :return:
    """
    files = {'file': open(path, 'rb')} if op is not OP.DELETE else None
    for item in SYNC_URLS:
        url = item.format(dag_id=dag_id, op=op.value)
        try:
            resp = requests.post(url, files=files, timeout=5)
            if resp.status_code != 200:
                logging.error('replicate {} failed, return: {} - {}', url, resp.status_code, resp.text)
                # todo 此处需要报警，并且考虑是否需要进一步的错误处理，例如回滚
        except Exception:
            logging.error('replicate {} failed, return: {} - {}', url, resp.status_code, resp.text)
            # todo 此处需要报警，并且考虑是否需要进一步的错误处理，例如回滚
    pass



# -*- coding: utf-8 -*-
# @Time    : 2019-04-11 17:33
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: cli.py
# @Software: PyCharm
import json
import os
import shutil
from zipfile import ZipFile

import requests


class DagManagerException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


class DMCParametersException(DagManagerException):
    pass


class DagManagerClient(object):
    __slots__ = ['base_url', 'folder']

    def __init__(self, base_url, folder):
        self.base_url = base_url
        self.folder = folder

    def get(self, dag_id=None):
        """
        从DagManager拉取指定的dag，dag_id为None时，拉取所有dag。
        会建一个和dag_id同名的文件夹存放dag文件，对于zip文件，会解压，并删除zip文件
        拉取指定dag，会先删除指定dag
        拉取所有dag时，会先删除所有dag
        :param dag_id:
        :return:
        """
        url = os.path.join(self.base_url, 'dags', dag_id or '')
        resp = requests.get(url)
        if resp.status_code != 200:
            raise DagManagerException(resp.text)
        d = os.path.join(self.folder, dag_id or '')
        # 拉取所有dag时，需先删除目前的所有的dag
        if dag_id is None:
            self._clear_folder()
        else:
            if os.path.exists(d):
                shutil.rmtree(d)
            os.mkdir(d)
        path = os.path.join(d, os.path.basename(resp.request.path_url))
        with open(path, 'wb') as f:
            f.write(resp.content)
        if resp.headers['Content-Type'] == 'application/zip':
            with ZipFile(path) as zf:
                zf.extractall(d)
            os.remove(path)
        # 拉取所有dag时，将解压后的所有文件放置在{self.folder}/{dag_id}/{dag_id}.[py|zip]
        # 对于zip的dag，还需要解压，并删除zip文件
        if dag_id is None:
            for filename in os.listdir(self.folder):
                old_path = os.path.join(self.folder, filename)
                new_dir = os.path.join(self.folder, os.path.splitext(filename)[0])
                new_path = os.path.join(new_dir, filename)
                os.mkdir(new_dir)
                os.rename(old_path, new_path)
                if filename.endswith('.zip'):
                    with ZipFile(new_path) as zf:
                        zf.extractall(new_dir)
                    os.remove(new_path)

    def _clear_folder(self):
        for d in os.listdir(self.folder):
            path = os.path.join(self.folder, d)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

    def remove(self, dag_id):
        path = os.path.join(self.folder, dag_id)
        if os.path.exists(path):
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)
        url = os.path.join(self.base_url, 'dags', dag_id or '')
        resp = requests.delete(url)
        if resp.status_code != 200:
            raise DagManagerException(resp.text)

    def clear(self):
        self._clear_folder()

    def list(self, pattern=None):
        url = os.path.join(self.base_url, 'dag_ids', pattern or '*')
        resp = requests.get(url)
        if resp.status_code != 200:
            raise DagManagerException(resp.text)
        dags = json.loads(resp.text)
        return sorted(dags)

    def upload(self, dag_id, filename, dag_path=None):
        url = os.path.join(self.base_url, 'dags', dag_id or '')
        files = {'file': open(filename, 'rb')}
        _, ext = os.path.splitext(filename)
        data = {}
        if ext == '.zip':
            if dag_path is None:
                raise DMCParametersException('must specify dag_path, when upload a zip file')
            data = {'dag_path': dag_path}
        resp = requests.post(url, files=files, data=data)
        if resp.status_code != 200:
            raise DagManagerException(resp.text)



# -*- coding: utf-8 -*-
# @Time    : 2019-04-15 19:42
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: setup.py
# @Software: PyCharm
from setuptools import find_packages, setup

setup(
    name='dag_manager',
    version='0.1.3a',
    author='wutong',
    author_email='twu@aibee.com',
    description='Manage airflow dags, synchronize updates to scheduler and webserver, '
    'so airflow does not depend on shared file system',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    zip_safe=False,
    url='http://gitlab.aibee.cn/platform/dag-manager',
    install_requires=[
        'flask==1.0.2',
        'requests==2.21.0',
        'redis==3.2.1',
        'apache-airflow==1.10.3',
    ],
    scripts=['dmc/bin/dmc']
)



# -*- coding: utf-8 -*-
# @Time    : 2019-04-15 14:49
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: cli.py
# @Software: PyCharm
import argparse
import json
import os
from collections import namedtuple

from dmc.dmc import DagManagerClient

Arg = namedtuple(
    'Arg', ['flags', 'help', 'action', 'default', 'nargs', 'type', 'choices', 'required', 'metavar'])
Arg.__new__.__defaults__ = (None, None, None, None, None, None, None)

with open(os.path.join(os.getenv('HOME'), '.mdc/config')) as f:
    conf = json.load(f)
base_url = conf['base_url']
storage_path = conf.get('storage_path') or os.getenv('MDC_DAG_HOME', os.getcwd())
dmc = DagManagerClient(base_url, storage_path)


def get(parsed_args):
    dmc.get(parsed_args.dag_id)


def getall(parsed_args):
    dmc.get()


def upload(parsed_args):
    dmc.upload(parsed_args.dag_id, parsed_args.file, parsed_args.dag_path)


def remove(parsed_args):
    dmc.remove(parsed_args.dag_id)


def ls(parsed_args):
    dag_ids = dmc.list(parsed_args.pattern)
    for dag_id in dag_ids:
        print(dag_id)


class CLIFactory(object):
    args = {
        'dag_id': Arg(('dag_id',), 'dag_id'),
        'file': Arg(('-f', '--file'), 'file to be uploaded'),
        'pattern': Arg(('-p', '--pattern'), 'pattern to be matched'),
        'dag_path': Arg(('-dp', '--dag_path'), 'path to the dag file, must specified when upload a zip file'),
    }
    subparsers = (
        {
            'func': ls,
            'help': 'list dags in DagManager by matching pattern',
            'args': ('pattern',)
        }, {
            'func': get,
            'help': 'get a dag from DagManager by dag_id',
            'args': ('dag_id', ),
        }, {
            'func': getall,
            'help': 'get all dags from DagManager',
            'args': tuple(),
        }, {
            'func': remove,
            'help': 'remove a dag from DagManager by dag_id',
            'args': ('dag_id', ),
        }, {
            'func': upload,
            'help': 'upload a dag to DagManager by dag_id',
            'args': ('dag_id', 'file', 'dag_path'),
        }
    )
    subparsers_dict = {sp['func'].__name__: sp for sp in subparsers}

    @classmethod
    def get_parser(cls):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(
            help='sub-command help', dest='subcommand')
        subparsers.required = True
        subparser_list = cls.subparsers_dict.keys()
        for sub in subparser_list:
            sub = cls.subparsers_dict[sub]
            sp = subparsers.add_parser(sub['func'].__name__, help=sub['help'])
            for arg in sub['args']:
                arg = cls.args[arg]
                kwargs = {
                    f: getattr(arg, f)
                    for f in arg._fields if f != 'flags' and getattr(arg, f)}
                sp.add_argument(*arg.flags, **kwargs)
            sp.set_defaults(func=sub['func'])
        return parser




# -*- coding: utf-8 -*-
# @Time    : 2019-04-16 11:56
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: config.py
# @Software: PyCharm

# DAG_FOLDER = '/user/local/airflow/dags'
DAG_FOLDER = 'static'



# -*- coding: utf-8 -*-
# @Time    : 2019-04-11 16:02
# @Author  : WuTong
# @Email   : twu@aibee.com
# @FileName: config.py
# @Software: PyCharm
import os

REDIS_CONFIG = {
    'host': 'bj-cpu008.aibee.cn',
    'port': 26379,
    'db': 3,
}


ALL_DAGS_ZIP_NAME = 'all_dags.zip'
DAG_FOLDER = 'app/static'
ALL_DAGS_ZIP_PATH = os.path.join(DAG_FOLDER, ALL_DAGS_ZIP_NAME)
ACCEPTED_EXT = ('.zip', '.py')
REDIS_PREFIX = 'D_M_PREFIX_'
REDIS_KEYS = {
    'lock_modification': REDIS_PREFIX + '___LocK_mODiFY__',
    'all_dag_ids': REDIS_PREFIX + '___dAgS___',
    'last_update_time': REDIS_PREFIX + '__lAsT_UpdAte_TiMe__',
    'last_zip_time': REDIS_PREFIX + '__lAsT_zIP_TiMe__',
}
# REDIS_ADD_DAGS = REDIS_PREFIX + '___AdD___dAgS___'
# REDIS_UPDATE_DAGS = REDIS_PREFIX + '___uPdAtE___dAgS___'
# REDIS_REMOVE_DAGS = REDIS_PREFIX + '___ReMoVe___dAgS___'


# 需同步更新的节点
SYNC_URLS = [
    "http://localhost:5001/{dag_id}/{op}"
]